#include "config.h"

#include <syslog.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <sys/uio.h>

#include <utlist.h>

#include "include/digitalSTROM/ds485-socket-server.h"
#include "include/digitalSTROM/ds485-socket-interface.h"
#include "ds485-socket-server-impl.h"

#include "base/log.h"
#include "base/ds_string.h"
#include "base/ds_packet.h"
#include "base/socket_send.h"

/* CAUTION: client->lock must be held */
int _ds485n_tlv_write(dS485dClient *cl, uint8_t cmd, const void *data, int dlen)
{
    uint16_t length = dlen + 3;
    struct iovec iov[2];
    uint8_t buf[8];
    int rval;

    buf[0] = cmd;
    buf[1] = length & 0xff;
    buf[2] = (length >> 8) & 0xff;

    iov[0].iov_base = buf;
    iov[0].iov_len = 3;
    iov[1].iov_base = (void *)data;
    iov[1].iov_len = dlen;

    if (cl->fd < 0) {
        return -EBADF;
    }

    rval = socket_sendv(cl->fd, iov, 2);
    if (rval) {
        report2(LOG_NOTICE, cl->peer_name, "_ds485n_tlv_write: cmd:0x%02x err:%s(%d)\n",
                cmd, strerror(-rval), -rval);
        _ds485n_client_detach(cl);
        return rval;
    }

    return 0;
}

/* CAUTION: client->lock must be held */
int _ds485n_dsmapi_write(dS485dClient* cl, uint8_t cmd,
                         const ds485_container_t *container)
{
    /* TODO see dS485CmdApidata, maybe remove it */
    return _ds485n_tlv_write(cl, cmd, container,
                             DS485_CONTAINER_SIZE + container->length);
}

NETLIB_EXPORTED void ds485_server_emit_bus_change(struct dS485_server *server, dsuid_t dsuid, int flags)
{
    dS485dClient* cs;
    ds485_container_t container;
    int rval;

    ds485n_create_bus_change(&container, &dsuid, flags);

    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {

        pthread_mutex_lock(&cs->lock);
        if (cs->fd < 0) {
            /* ignore leaving clients */
            pthread_mutex_unlock(&cs->lock);
            continue;
        }

        rval = _ds485n_dsmapi_write(cs, DS485D_CMD_BUS_CHANGE, &container);
        pthread_mutex_unlock(&cs->lock);

        if (rval) {
            report2(LOG_NOTICE, cs->peer_name, "missed bus change err:%s(%d)\n",
                    strerror(-rval), -rval);
            continue;
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);
}

NETLIB_EXPORTED void ds485_server_emit_bus_state(struct dS485_server *server, bus_state_t state)
{
    dS485dClient* cs;
    ds485_container_t container;
    int rval;

    ds485n_create_bus_state(&container, state);

    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {

        pthread_mutex_lock(&cs->lock);
        if (cs->fd < 0) {
            /* ignore leaving clients */
            pthread_mutex_unlock(&cs->lock);
            continue;
        }

        rval = _ds485n_dsmapi_write(cs, DS485D_CMD_BUS_STATE, &container);
        pthread_mutex_unlock(&cs->lock);

        if (rval) {
            report2(LOG_NOTICE, cs->peer_name, "missed state change err:%s(%d)\n",
                    strerror(-rval), -rval);
            continue;
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);
}

/**
 * Sends packet to client
 *
 * Problems of the routing protocol ds485-netlib v1.
 * - clients do not have an address
 * - netlib assigns clients a transactionId upon connect
 *   and sets source address to be the address of the server
 *
 * The transactionId is abused to tag the container with the
 * client connection. A reply is only sent to the client with the
 * matching transactionId.
 * In a ds485p / ds485 stacked setup. 1st the ds485p changes the
 * transactionId of the container, then ds485d again. So the
 * ds485p transactionId is lost, and ds485p can't rely on it.
 *
 * The solution to this was to split the transaction id into
 * halves, ds485p is using th upper 4 bits, vdsm/ds485d the lower
 * 4 bits. This is the reason why you can't connect a ds485p to
 * another ds485p, they would both overwrite the same
 * transactionId and replies would be lost
 *
 * packet is forwarded
 * - to client with matching transactionId
 * - all sniffing clients
 * - all clients if message is event or request,
 *   (cache-snooping)
 */
static void ds485n_pu_send(struct dS485_server *server,
                           const packet_union_t *packet)
{
    dS485dClient* cs;
    PKT_UNION_STR_BUF(buf);
    bool delivered = false;
    int forwardId;
    int rval;

    if (ds485b_pu_is_oversize(packet)) {
        sprintf_packet_union(buf, packet);
        report(LOG_WARNING, "send: oversize packet: %s\n", buf);
        return;
    }

    forwardId = (server->upperPath)
        ? packet->transactionId >> 4
        : packet->transactionId & 0x0f;

    server->rxPacketNum++;

    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {
        /*
         * matching request and response is a hard problem
         * ds485-client inspects cmd/modifier in container
         * data. Problem if another client issued an
         * identical request to the same dsm, --
         * but different target device, whose id is part of
         * container data, -- the ds485-client can't
         * distinguish the replies anymore, it would have
         * to decode the complete container data, which
         * differs for each command.
         * Hence it's important to forward replies only to
         * the right client. That said, setting
         * DS485_FILTER_ALL_RESPONSES exludes clients from
         * issueing synchronous request, they are limited
         * to logging/snooping traffic
         */
        if ((cs->transactionId == forwardId) ||
            (packet->containerType != DS485_CONTAINER_RESPONSE) ||
            (cs->rxfilter & DS485_FILTER_ALL_RESPONSES)) {

            pthread_mutex_lock(&cs->lock);
            if (cs->fd < 0) {
                /* ignore leaving clients */
                pthread_mutex_unlock(&cs->lock);
                continue;
            }

            rval = _ds485n_pu_write(cs, packet);
            pthread_mutex_unlock(&cs->lock);

            if (rval) {
                sprintf_packet_union(buf, packet);
                report2(LOG_NOTICE, cs->peer_name, "send err:%s(%d) pkt:%s\n",
                        strerror(-rval), -rval, buf);
                continue;
            }

            server->txPacketNum++;
            delivered |= (cs->transactionId == forwardId);
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);

    /*
     * Suppress warning for request containers. client requests
     * are forwarded to all other clients. Imagine a
     * ds485d/ds485 setup with an additionnal ds485d client
     * doing reqeusts. ds485d will forward to ds485p, ds485p
     * send to all its client, here we are -- of course the
     * destination is not found, unless the the transaction
     * id accidently matched
     * TODO: likely transaction id is null, since ds485d
     * modifies the lower 4bits, and transaction id '0' is
     * reserved. Unless there is another ds485p running in
     * parallel
     * BETTER: flag the container itself as
     * coherence-request
     */
    if (!delivered && !dsuid_is_broadcast(&packet->dst) &&
        (packet->containerType != DS485_CONTAINER_REQUEST)) {
        sprintf_packet_union(buf, packet);
        report(LOG_NOTICE, "send undelivered: %s\n", buf);
    }
}

NETLIB_EXPORTED void ds485_server_reply(struct dS485_server *server,
                        const ds485_container_t *container)
{
    const packet_union_t data;
    ds485b_pu_from_short((packet_union_t *)&data, container);
    return ds485n_pu_send(server, &data);
}

NETLIB_EXPORTED void ds485_server_reply_long(struct dS485_server *server,
                             const ds485n_packet_t *lcont)
{
    const packet_union_t data;
    ds485b_pu_from_long((packet_union_t *)&data, lcont);
    return ds485n_pu_send(server, &data);
}

/**
 * Forward packet to server and all peers incl. ourselves if promiscous
 */
int ds485n_pu_receive(struct dS485_server *server, packet_union_t *packet,
                      dS485dClient *sender)
{
    PKT_UNION_STR_BUF(buf);
    dS485dClient* cs;
    int ret;

    /* netlib_packet_parser always passes sender */
    assert(sender);

    if (ds485b_pu_is_oversize(packet)) {
        sprintf_packet_union(buf, packet);
        report(LOG_WARNING, "receive: oversize packet: %s\n", buf);
        return -ENOBUFS;
    }

    if (server->upperPath) {
        packet->transactionId &= 0x0f;
        packet->transactionId |= (sender->transactionId << 4);
    } else {
        packet->transactionId &= 0xf0;
        packet->transactionId |= (sender->transactionId & 0x0f);
    }

    // ds485c has no dsuid
    if (dsuid_is_null(&packet->src)) {
        packet->src = server->cb.get_dsuid(server);
    }

    ds485b_pu_sync(packet);

    sender->m_state = CL_SERVER_RECEIVE;
    ret = (packet->type == v_container) ?
        server->cb.receive_data(server, packet->u.container) :
        server->cb.receive_long_data(server, packet->u.packet);
    if (ret) {
        sprintf_packet_union(buf, packet);
        report(LOG_NOTICE, "receive: srv err:%s(%d) bus:%s pkt:%s\n",
               strerror(-ret), -ret, sender->peer_name, buf);
        /* reinvent ICMP return ret;  TODO */
    }

    sender->m_state = CL_CACHE_SNOOP_START;
    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {
        int rval;
        /*
         * Poor man's cache-snooping, if this client
         * is doing some configuration change, give
         * all other clients a heads up
         * In a ds485d<->client environment, sending a packet
         * automatically meant sending a request-container.
         * With a ds485p in between we could also be
         * forwarding a response/event from one bus to
         * another. Currently those are not forwarded,
         * but filtered out in the ds485p
         */
        if ((cs != sender) || (cs->rxfilter & DS485_FILTER_ALL_RESPONSES)) {

            sender->m_state = CL_CACHE_SNOOP_BUDDY_LOCK;
            pthread_mutex_lock(&cs->lock);
            if (cs->fd < 0) {
                /* ignore leaving clients */
                pthread_mutex_unlock(&cs->lock);
                continue;
            }

            rval = _ds485n_pu_write(cs, packet);
            pthread_mutex_unlock(&cs->lock);

            if (rval) {
                sprintf_packet_union(buf, packet);
                report2(LOG_NOTICE, cs->peer_name,
                        "receive: cl err:%s(%d) bus:%s pkt:%s\n",
                        strerror(-rval), -rval, sender->peer_name, buf);
                continue;
            }
            server->txPacketNum++;
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);
    return ret;
}
